@@ -11,7 +11,7 @@
 #endif /* HAVE_CONFIG_H */
 
 #ifdef AUTH_PAM
-static const char rcsid[] = "$Id: auth_pam.c,v 1.14 2001/04/23 23:45:40 chris Exp $";
+static const char rcsid[] = "$Id: auth_pam.c,v 1.15 2001/06/10 21:52:10 chris Exp $";
 
 #include <sys/types.h> /* BSD needs this here, it seems. */
 
@@ -70,7 +70,15 @@ authcontext auth_pam_new_user_pass(const char *user, const char *pass) {
     item *I;
     int use_gid = 0;
     gid_t gid = 99;
+    const char *x;
 
+    /* Check the this isn't a virtual-domain user. */
+    x = user + strcspn(user, "@%!");
+    if (*x) return NULL;
+
+    /* Copy the password structure, since it is in static storage and may
+     * get overwritten by calls in the PAM code.
+     */
     pw2 = getpwnam(user);
     if (!pw2) return NULL;
     else memcpy(&pw, pw2, sizeof(pw));
