@@ -11,7 +11,7 @@
 #endif /* HAVE_CONFIG_H */
 
 #ifdef AUTH_PASSWD
-static const char rcsid[] = "$Id: auth_passwd.c,v 1.8 2001/04/23 23:45:40 chris Exp $";
+static const char rcsid[] = "$Id: auth_passwd.c,v 1.9 2001/06/10 21:52:10 chris Exp $";
 
 #include <sys/types.h>
 
@@ -56,6 +56,11 @@ authcontext auth_passwd_new_user_pass(const char *user, const char *pass) {
     int use_gid = 0;
     gid_t gid = 99;
     authcontext a = NULL;
+    const char *x;
+
+    /* Check the this isn't a virtual-domain user. */
+    x = user + strcspn(user, "@%!");
+    if (*x) return NULL;
 
     pw = getpwnam(user);
     if (!pw) return NULL;
