@@ -4,13 +4,17 @@
  *
  * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
  *
- * $Id: auth_passwd.h,v 1.4 2001/01/11 22:28:08 chris Exp $
+ * $Id: auth_passwd.h,v 1.5 2001/04/23 23:45:40 chris Exp $
  *
  */
 
 #ifndef __AUTH_PASSWD_H_ /* include guard */
 #define __AUTH_PASSWD_H_
 
+#ifdef HAVE_CONFIG_H
+#include "configuration.h"
+#endif /* HAVE_CONFIG_H */
+
 #ifdef AUTH_PASSWD
 
 #include "authswitch.h"
