@@ -1,13 +1,13 @@
 /*
  * main.c:
- * main loop for pop3 server
+ * Entry point, initialisation code and main loop for pop3 server.
  *
  * Copyright (c) 2001 Chris Lightfoot. All rights reserved.
  *
  */
 
 static const char copyright[] = "$Copyright: (c) 2001 Chris Lightfoot. All rights reserved. $";
-static const char rcsid[] = "$Id: main.c,v 1.42 2001/04/26 16:30:03 chris Exp $";
+static const char rcsid[] = "$Id: main.c,v 1.48 2001/06/10 21:52:10 chris Exp $";
 
 #ifdef HAVE_CONFIG_H
 #include "configuration.h"
@@ -72,6 +72,7 @@ connection this_child_connection;   /* Stored here so that if a signal terminate
 vector listeners;                   /* Listeners */
 list connections;                   /* Active connections. */
 
+
 /* Theory of operation:
  * The main loop is in net_loop, below; it calls listeners_ and
  * connections_pre_select, then calls select, then calls listeners_ and
@@ -80,7 +81,6 @@ list connections;                   /* Active connections. */
  * connections are used to handle this procedure.
  */
 
-
 /* listeners_pre_select:
  * Called before the main select(2) so listening sockets can be polled.
  */
@@ -110,12 +110,13 @@ void listeners_post_select(fd_set *readfds, fd_set *writefds, fd_set *exceptfds)
             if (s == -1) {
                 if (errno != EAGAIN) print_log(LOG_ERR, "net_loop: accept: %m");
             } else if (setsockopt(s, SOL_SOCKET, SO_SNDBUF, &a, sizeof(a)) == -1) {
-                /* Set a small send buffer so that we get usefully blocking writes. */
-                print_log(LOG_ERR, "handle_listeners: setsockopt: %m");
+                /* Set a small send buffer so that we get usefully blocking
+                 * writes. */
+                print_log(LOG_ERR, "listeners_post_select: setsockopt: %m");
                 close(s);
             } else if (fcntl(s, F_SETFL, 0) == -1) {
                 /* Switch off non-blocking mode, in case it is inherited. */
-                print_log(LOG_ERR, "handle_listeners: fcntl(F_SETFL): %m");
+                print_log(LOG_ERR, "listeners_post_select: fcntl(F_SETFL): %m");
                 close(s);
             } else {
                 if (num_running_children >= max_running_children) {
@@ -123,11 +124,15 @@ void listeners_post_select(fd_set *readfds, fd_set *writefds, fd_set *exceptfds)
                     xwrite(s, m, strlen(m));
                     shutdown(s, 2);
                     close(s);
-                    print_log(LOG_INFO, _("handle_listeners: rejected connection from %s owing to high load"), inet_ntoa(sin.sin_addr));
+                    print_log(LOG_INFO, _("listeners_post_select: rejected connection from %s owing to high load"), inet_ntoa(sin.sin_addr));
                 } else {
                     connection c = connection_new(s, &sin, L->domain);
-                    if (c) list_push_back(connections, item_ptr(c));
-                    print_log(LOG_INFO, _("handle_listeners: client %s: connected"), c->idstr);
+                    if (c) {
+                        list_push_back(connections, item_ptr(c));
+                        print_log(LOG_INFO, _("listeners_post_select: client %s: connected"), c->idstr);
+                    } else
+                        /* This could be really bad, but all we can do is log the failure. */
+                        print_log(LOG_ERR, _("listeners_post_select: unable to set up connection from %s: %m"), inet_ntoa(sin.sin_addr));
                 }
             }
         }
@@ -157,6 +162,7 @@ void fork_child(connection *C, listitem *i) {
     listitem J;
     item *t;
     post_fork = 1; /* This is right. See below. */
+    muntrace();
     switch(fork()) {
         case 0:
             /* Child. Dispose of listeners and connections other than this
@@ -180,7 +186,7 @@ void fork_child(connection *C, listitem *i) {
                 print_log(LOG_ERR, _("fork_child: client %s: authentication context has UID of 0"), c->idstr);
                 connection_sendresponse(c, 0, _("Everything's really bad"));
                 connection_delete(c);
-                exit(0);
+                _exit(0);
             }
 
             /* Set our gid and uid to that appropriate for the mailspool, as decided by the auth switch. */
@@ -188,24 +194,39 @@ void fork_child(connection *C, listitem *i) {
                 print_log(LOG_ERR, "fork_child: setgid(%d): %m", c->a->gid);
                 connection_sendresponse(c, 0, _("Something bad happened, and I just can't go on. Sorry."));
                 connection_delete(c);
-                exit(0);
+                _exit(0);
             } else if (setuid(c->a->uid) == -1) {
                 print_log(LOG_ERR, "fork_child: setuid(%d): %m", c->a->uid);
                 connection_sendresponse(c, 0, _("Something bad happened, and I just can't go on. Sorry."));
                 connection_delete(c);
-                exit(0);
+                _exit(0);
             }
 
             /* Get in to the `transaction' state, opening the mailbox. */
             if (connection_start_transaction(c)) {
-                char s[1024];
-                snprintf(s, 1024, _("Welcome aboard! You have %d messages."), c->m->index->n_used);
+                char s[512], *p;
+                strcpy(s, _("Welcome aboard!"));
+                strcat(s, " ");
+                p = s + strlen(s);
+                switch (c->m->index->n_used) {
+                    case 0:
+                        strcpy(p, _("You have no messages at all."));
+                        break;
+
+                    case 1:
+                        strcat(p, _("You have exactly one message."));
+                        break;
+
+                    default:
+                        sprintf(p, _("You have %d messages."), c->m->index->n_used);
+                        break;
+                }
                 connection_sendresponse(c, 1, s);
                 this_child_connection = c;
             } else {
                 connection_sendresponse(c, 0, _("Unable to open mailbox; it may be locked by another concurrent session."));
                 connection_delete(c);
-                exit(0);
+                _exit(0);
             }
 
             I = NULL;
@@ -254,25 +275,27 @@ void connections_post_select(fd_set *readfds, fd_set *writefds, fd_set *exceptfd
                 print_log(LOG_INFO, _("connections_post_select: connection_read: client %s: closed connection"), c->idstr);
                 connection_delete(c);
                 I = list_remove(connections, I);
-                if (post_fork) exit(0);
+                if (post_fork) _exit(0);
                 if (!I) break;
             } else if (n < 0 && errno != EINTR) {
                 /* Some sort of error occurred, and we should close the connection. */
                 print_log(LOG_ERR, _("connections_post_select: connection_read: client %s: disconnected: %m"), c->idstr);
                 connection_delete(c);
                 I = list_remove(connections, I);
-                if (post_fork) exit(0);
+                if (post_fork) _exit(0);
                 if (!I) break;
             } else {
                 /* We read some data and should try to interpret command/s. */
                 pop3command p;
                 while (c && (p = connection_parsecommand(c))) {
-                    switch(connection_do(c, p)) {
+                    enum connection_action act = connection_do(c, p);
+                    pop3command_delete(p);
+                    switch (act) {
                         case close_connection:
                             connection_delete(c);
                             c = NULL;
                             I = list_remove(connections, I);
-                            if (post_fork) exit(0);
+                            if (post_fork) _exit(0);
                             break;
 
                         case fork_and_setuid:
@@ -288,8 +311,6 @@ void connections_post_select(fd_set *readfds, fd_set *writefds, fd_set *exceptfd
                         default:;
                     }
 
-                    pop3command_delete(p);
-
                     if (!I) break;
                 }
 
@@ -304,7 +325,7 @@ void connections_post_select(fd_set *readfds, fd_set *writefds, fd_set *exceptfd
 #endif
             print_log(LOG_INFO, "net_loop: timed out client %s", ((connection)I->d.v)->idstr);
             connection_delete((connection)(I->d.v));
-            if (post_fork) exit(0);
+            if (post_fork) _exit(0);
             I = list_remove(connections, I);
             if (!I) break;
         }
@@ -317,8 +338,18 @@ void connections_post_select(fd_set *readfds, fd_set *writefds, fd_set *exceptfd
  */
 volatile int foad = 0, restart = 0; /* Flags used to indicate that we should exit or should re-exec. */
 
+#ifdef AUTH_OTHER
+extern pid_t authchild_died;
+extern int authchild_status;
+#endif /* AUTH_OTHER */
+
 void net_loop() {
     listitem J;
+#ifdef AUTH_OTHER
+    sigset_t chmask;
+    sigemptyset(&chmask);
+    sigaddset(&chmask, SIGCHLD);
+#endif /* AUTH_OTHER */
 
     connections = list_new();
 
@@ -346,10 +377,26 @@ void net_loop() {
             /* Monitor existing connections */
             connections_post_select(&readfds, NULL, NULL);
         }
+
+#ifdef AUTH_OTHER
+        /* It may be that the authentication child died; log the message here
+         * to avoid doing something we shouldn't in the signal handler. We
+         * block SIGCHLD while doing this.
+         */
+        sigprocmask(SIG_BLOCK, &chmask, NULL);
+        if (authchild_died) {
+            print_log(LOG_WARNING, _("net_loop: authentication child %d terminated with status %d"), (int)authchild_died, authchild_status);
+            authchild_died = 0;
+        }
+        sigprocmask(SIG_UNBLOCK, &chmask, NULL);
+#endif /* AUTH_OTHER */
     }
 
-    /* Termination request received; we should close all connections in an orderly fashion. */
-    if (restart) print_log(LOG_INFO, _("net_loop: restarting on signal %d"), foad);
+    /* Termination request received; we should close all connections in an
+     * orderly fashion.
+     */
+    if (restart) print_log(LOG_INFO, _("net_loop: restarting on signal %d"),
+            foad);
     else print_log(LOG_INFO, _("net_loop: terminating on signal %d"), foad);
 
     if (connections) {
@@ -405,6 +452,8 @@ int main(int argc, char **argv, char **envp) {
     char *configfile = "/etc/tpop3d.conf", c;
     int na;
 
+    mtrace();
+
     /* Read the options. */
     opterr = 0;
     while ((c = getopt(argc, argv, optstring)) != EOF) {
@@ -432,10 +481,8 @@ int main(int argc, char **argv, char **envp) {
 
             case '?':
             default:
-                if (optopt == 'f' && !optarg)
-                    fprintf(stderr, _("tpop3d: option -f requires an argument\n"));
-                else if (optopt == 'p' && !optarg)
-                    fprintf(stderr, _("tpop3d: option -p requires an argument\n"));
+                if (strchr(optstring, optopt))
+                    fprintf(stderr, _("tpop3d: option -%c requires an argument\n"), optopt);
                 else
                     fprintf(stderr, _("tpop3d: unrecognised option -%c\n"), optopt);
                 usage(stderr);
@@ -565,6 +612,7 @@ int main(int argc, char **argv, char **envp) {
                 print_log(LOG_ERR, _("%s: value for max-children must be 1 or greater; exiting"), configfile);
                 EXIT_REMOVING_PIDFILE(1);
             }
+            break;
 
         default:
             max_running_children = 16;
@@ -585,6 +633,7 @@ int main(int argc, char **argv, char **envp) {
                 print_log(LOG_ERR, _("%s: value for timeout-seconds must be 1 or greater; exiting"), configfile);
                 EXIT_REMOVING_PIDFILE(1);
             }
+            break;
 
         default:
             timeout_seconds = 30;
@@ -617,18 +666,20 @@ int main(int argc, char **argv, char **envp) {
     EXIT_REMOVING_PIDFILE(0);
 }
 
+#if 0
+/* Primitive memory-leak debugging. */
 #undef malloc
 #undef free
+#undef realloc
 
-#if 0
-char *mystrdup(char *f, int l, char *s) {
+char *mystrdup(char *f, int l, const char *s) {
     char *p = malloc(strlen(s) + 1);
     strcpy(p, s);
     fprintf(stderr, "[%d] %s:%d: %p = strdup(\"%s\")\n", getpid(), f, l, p, s);
     return p;
 }
 
-void *mymalloc(char *f, int l, size_t n) {
+void *mymalloc(char *f, int l, const size_t n) {
     void *p = malloc(n);
     if (!p) return NULL;
     fprintf(stderr, "[%d] %s:%d: %p = malloc(%d)\n", getpid(), f, l, p, n);
@@ -639,4 +690,12 @@ void myfree(char *f, int l, void *p) {
     free(p);
     fprintf(stderr, "[%d] %s:%d: free(%p)\n", getpid(), f, l, p);
 }
+
+void *myrealloc(char *f, int l, void *p, const size_t n) {
+    void *q = realloc(p, n);
+    fprintf(stderr, "[%d] %s:%d: free(%p)          (actually realloc)\n", getpid(), f, l, p);
+    fprintf(stderr, "[%d] %s:%d: malloc(%p, %d)    (actually realloc)\n", getpid(), f, l, p, n);
+    return q;
+}
+
 #endif 
